@@ -1,3 +1,27 @@
+"""
+MIT License
+
+Copyright (c) 2021 Alex Hall
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
+
 import __future__
 import ast
 import dis
@@ -5,13 +29,17 @@ import functools
 import inspect
 import io
 import linecache
+import re
 import sys
 import types
 from collections import defaultdict, namedtuple
+from copy import deepcopy
 from itertools import islice
 from operator import attrgetter
 from threading import RLock
 
+function_node_types = (ast.FunctionDef,)
+
 PY3 = sys.version_info[0] == 3
 
 if PY3:
@@ -49,6 +77,7 @@ else:
 
     # noinspection PyUnresolvedReferences
     text_type = unicode
+
 try:
     # noinspection PyUnresolvedReferences
     _get_instructions = dis.get_instructions
@@ -57,7 +86,7 @@ except AttributeError:
         lineno = None
 
 
-    from dis import HAVE_ARGUMENT, EXTENDED_ARG, hasconst, opname, findlinestarts
+    from dis import HAVE_ARGUMENT, EXTENDED_ARG, hasconst, opname, findlinestarts, hasname
 
     # Based on dis.disassemble from 2.7
     # Left as similar as possible for easy diff
@@ -84,9 +113,19 @@ except AttributeError:
 
                 if op in hasconst:
                     argval = co.co_consts[oparg]
+                elif op in hasname:
+                    argval = co.co_names[oparg]
+                elif opname[op] == 'LOAD_FAST':
+                    argval = co.co_varnames[oparg]
             yield Instruction(offset, argval, opname[op], lineno)
 
 
+try:
+    function_node_types += (ast.AsyncFunctionDef,)
+except AttributeError:
+    pass
+
+
 def assert_(condition, message=""):
     """
     Like an assert statement, but unaffected by -O
@@ -98,7 +137,7 @@ def assert_(condition, message=""):
 
 
 def get_instructions(co):
-    lineno = None
+    lineno = co.co_firstlineno
     for inst in _get_instructions(co):
         lineno = inst.starts_line or lineno
         assert_(lineno)
@@ -193,8 +232,13 @@ class Source(object):
             for node in ast.walk(self.tree):
                 for child in ast.iter_child_nodes(node):
                     child.parent = node
-                if hasattr(node, 'lineno'):
-                    self._nodes_by_line[node.lineno].append(node)
+                if hasattr(node, "lineno"):
+                    if hasattr(node, "end_lineno") and isinstance(node, ast.expr):
+                        linenos = range(node.lineno, node.end_lineno + 1)
+                    else:
+                        linenos = [node.lineno]
+                    for lineno in linenos:
+                        self._nodes_by_line[lineno].append(node)
 
             visitor = QualnameVisitor()
             visitor.visit(self.tree)
@@ -273,15 +317,20 @@ class Source(object):
             args = executing_cache[key]
         except KeyError:
             def find(source, retry_cache):
-                node = stmts = None
+                node = stmts = decorator = None
                 tree = source.tree
                 if tree:
                     try:
                         stmts = source.statements_at_line(lineno)
                         if stmts:
-                            if code.co_filename.startswith('<ipython-input-') and code.co_name == '<module>':
+                            if code.co_name == "<module>" and re.search(
+                                r"<ipython-input-|[/\\]ipykernel_\d+[/\\]",
+                                code.co_filename,
+                            ):
                                 tree = _extract_ipython_statement(stmts, tree)
-                            node = NodeFinder(frame, stmts, tree, lasti).result
+                            node_finder = NodeFinder(frame, stmts, tree, lasti)
+                            node = node_finder.result
+                            decorator = node_finder.decorator
                     except Exception as e:
                         # These exceptions can be caused by the source code having changed
                         # so the cached Source doesn't match the running code
@@ -300,7 +349,7 @@ class Source(object):
                         assert_(new_stmts <= stmts)
                         stmts = new_stmts
 
-                return source, node, stmts
+                return source, node, stmts, decorator
 
             args = find(source=cls.for_frame(frame), retry_cache=True)
             executing_cache[key] = args
@@ -395,13 +444,19 @@ class Executing(object):
 
     Generally you will just want `node`, which is the AST node being executed,
     or None if it's unknown.
+
+    If a decorator is currently being called, then:
+        - `node` is a function or class definition
+        - `decorator` is the expression in `node.decorator_list` being called
+        - `statements == {node}`
     """
 
-    def __init__(self, frame, source, node, stmts):
+    def __init__(self, frame, source, node, stmts, decorator):
         self.frame = frame
         self.source = source
         self.node = node
         self.statements = stmts
+        self.decorator = decorator
 
     def code_qualname(self):
         return self.source.code_qualname(self.frame.f_code)
@@ -419,7 +474,7 @@ class QualnameVisitor(ast.NodeVisitor):
         self.stack = []
         self.qualnames = {}
 
-    def visit_FunctionDef(self, node, name=None):
+    def add_qualname(self, node, name=None):
         name = name or node.name
         self.stack.append(name)
         if getattr(node, 'decorator_list', ()):
@@ -428,6 +483,8 @@ class QualnameVisitor(ast.NodeVisitor):
             lineno = node.lineno
         self.qualnames.setdefault((name, lineno), ".".join(self.stack))
 
+    def visit_FunctionDef(self, node, name=None):
+        self.add_qualname(node, name)
         self.stack.append('<locals>')
         if isinstance(node, ast.Lambda):
             children = [node.body]
@@ -451,12 +508,14 @@ class QualnameVisitor(ast.NodeVisitor):
                     if isinstance(grandchild, ast.AST):
                         self.visit(grandchild)
 
+    visit_AsyncFunctionDef = visit_FunctionDef
+
     def visit_Lambda(self, node):
         # noinspection PyTypeChecker
         self.visit_FunctionDef(node, '<lambda>')
 
     def visit_ClassDef(self, node):
-        self.stack.append(node.name)
+        self.add_qualname(node)
         self.generic_visit(node)
         self.stack.pop()
 
@@ -482,6 +541,7 @@ sentinel = 'io8urthglkjdghvljusketgIYRFYUVGHFRTBGVHKGF78678957647698'
 
 class NodeFinder(object):
     def __init__(self, frame, stmts, tree, lasti):
+        assert_(stmts)
         self.frame = frame
         self.tree = tree
         self.code = code = frame.f_code
@@ -496,9 +556,11 @@ class NodeFinder(object):
         else:
             self.ignore_linenos = frozenset()
 
-        instruction = self.get_actual_current_instruction(lasti)
+        self.decorator = None
+
+        self.instruction = instruction = self.get_actual_current_instruction(lasti)
         op_name = instruction.opname
-        self.lasti = instruction.offset
+        extra_filter = lambda e: True
 
         if op_name.startswith('CALL_'):
             typ = ast.Call
@@ -506,12 +568,42 @@ class NodeFinder(object):
             typ = ast.Subscript
         elif op_name.startswith('BINARY_'):
             typ = ast.BinOp
+            op_type = dict(
+                BINARY_POWER=ast.Pow,
+                BINARY_MULTIPLY=ast.Mult,
+                BINARY_MATRIX_MULTIPLY=getattr(ast, "MatMult", ()),
+                BINARY_FLOOR_DIVIDE=ast.FloorDiv,
+                BINARY_TRUE_DIVIDE=ast.Div,
+                BINARY_MODULO=ast.Mod,
+                BINARY_ADD=ast.Add,
+                BINARY_SUBTRACT=ast.Sub,
+                BINARY_LSHIFT=ast.LShift,
+                BINARY_RSHIFT=ast.RShift,
+                BINARY_AND=ast.BitAnd,
+                BINARY_XOR=ast.BitXor,
+                BINARY_OR=ast.BitOr,
+            )[op_name]
+            extra_filter = lambda e: isinstance(e.op, op_type)
         elif op_name.startswith('UNARY_'):
             typ = ast.UnaryOp
+            op_type = dict(
+                UNARY_POSITIVE=ast.UAdd,
+                UNARY_NEGATIVE=ast.USub,
+                UNARY_NOT=ast.Not,
+                UNARY_INVERT=ast.Invert,
+            )[op_name]
+            extra_filter = lambda e: isinstance(e.op, op_type)
         elif op_name in ('LOAD_ATTR', 'LOAD_METHOD', 'LOOKUP_METHOD'):
             typ = ast.Attribute
+            # `in` to handle private mangled attributes
+            extra_filter = lambda e: e.attr in instruction.argval
+        elif op_name in ('LOAD_NAME', 'LOAD_GLOBAL', 'LOAD_FAST', 'LOAD_DEREF', 'LOAD_CLASSDEREF'):
+            typ = ast.Name
+            if PY3 or instruction.argval:
+                extra_filter = lambda e: e.id == instruction.argval
         elif op_name in ('COMPARE_OP', 'IS_OP', 'CONTAINS_OP'):
             typ = ast.Compare
+            extra_filter = lambda e: len(e.ops) == 1
         else:
             raise RuntimeError(op_name)
 
@@ -522,15 +614,53 @@ class NodeFinder(object):
                 for node in ast.walk(stmt)
                 if isinstance(node, typ)
                 if not (hasattr(node, "ctx") and not isinstance(node.ctx, ast.Load))
+                if extra_filter(node)
+                if statement_containing_node(node) == stmt
             }
 
-            self.result = only(list(self.matching_nodes(exprs)))
+            matching = list(self.matching_nodes(exprs))
+            if not matching and typ == ast.Call:
+                self.find_decorator(stmts)
+            else:
+                self.result = only(matching)
+
+    def find_decorator(self, stmts):
+        stmt = only(stmts)
+        assert_(isinstance(stmt, (ast.ClassDef, function_node_types)))
+        decorators = stmt.decorator_list
+        assert_(decorators)
+        line_instructions = [
+            inst
+            for inst in self.clean_instructions(self.code)
+            if inst.lineno == self.frame.f_lineno
+        ]
+        last_decorator_instruction_index = [
+            i
+            for i, inst in enumerate(line_instructions)
+            if inst.opname == "CALL_FUNCTION"
+        ][-1]
+        assert_(
+            line_instructions[last_decorator_instruction_index + 1].opname.startswith(
+                "STORE_"
+            )
+        )
+        decorator_instructions = line_instructions[
+            last_decorator_instruction_index
+            - len(decorators)
+            + 1 : last_decorator_instruction_index
+            + 1
+        ]
+        assert_({inst.opname for inst in decorator_instructions} == {"CALL_FUNCTION"})
+        decorator_index = decorator_instructions.index(self.instruction)
+        decorator = decorators[::-1][decorator_index]
+        self.decorator = decorator
+        self.result = stmt
 
     def clean_instructions(self, code):
         return [
             inst
             for inst in get_instructions(code)
-            if inst.opname != 'EXTENDED_ARG'
+            if inst.opname not in ("EXTENDED_ARG", "NOP")
             if inst.lineno not in self.ignore_linenos
         ]
 
@@ -557,9 +687,9 @@ class NodeFinder(object):
         original_index = only(
             i
             for i, inst in enumerate(original_instructions)
-            if inst.offset == self.lasti
+            if inst == self.instruction
         )
-        for i, expr in enumerate(exprs):
+        for expr_index, expr in enumerate(exprs):
             setter = get_setter(expr)
             # noinspection PyArgumentList
             replacement = ast.BinOp(
@@ -573,6 +703,16 @@ class NodeFinder(object):
                 instructions = self.compile_instructions()
             finally:
                 setter(expr)
+
+            if sys.version_info >= (3, 10):
+                try:
+                    handle_jumps(instructions, original_instructions)
+                except Exception:
+                    # Give other candidates a chance
+                    if TESTING or expr_index < len(exprs) - 1:
+                        continue
+                    raise
+
             indices = [
                 i
                 for i, instruction in enumerate(instructions)
@@ -614,31 +754,12 @@ class NodeFinder(object):
                     instructions.pop(new_index + 1)
 
                 # Check that the modified instructions don't have anything unexpected
-                for inst1, inst2 in zip_longest(original_instructions, instructions):
-                    assert_(
-                        inst1.opname == inst2.opname or
-                        all(
-                            'JUMP_IF_' in inst.opname
-                            for inst in [inst1, inst2]
-                        ) or
-                        all(
-                            inst.opname in ('JUMP_FORWARD', 'JUMP_ABSOLUTE')
-                            for inst in [inst1, inst2]
-                        )
-                        or (
-                                inst1.opname == 'PRINT_EXPR' and
-                                inst2.opname == 'POP_TOP'
-                        )
-                        or (
-                                inst1.opname in ('LOAD_METHOD', 'LOOKUP_METHOD') and
-                                inst2.opname == 'LOAD_ATTR'
-                        )
-                        or (
-                                inst1.opname == 'CALL_METHOD' and
-                                inst2.opname == 'CALL_FUNCTION'
-                        ),
-                        (inst1, inst2, ast.dump(expr), expr.lineno, self.code.co_filename)
-                    )
+                # 3.10 is a bit too weird to assert this in all cases but things still work
+                if sys.version_info < (3, 10):
+                    for inst1, inst2 in zip_longest(
+                        original_instructions, instructions
+                    ):
+                        assert_(inst1 and inst2 and opnames_match(inst1, inst2))
 
                 yield expr
 
@@ -704,6 +825,142 @@ class NodeFinder(object):
             index += 1
 
 
+def non_sentinel_instructions(instructions, start):
+    """
+    Yields (index, instruction) pairs excluding the basic
+    instructions introduced by the sentinel transformation
+    """
+    skip_power = False
+    for i, inst in islice(enumerate(instructions), start, None):
+        if inst.argval == sentinel:
+            assert_(inst.opname == "LOAD_CONST")
+            skip_power = True
+            continue
+        elif skip_power:
+            assert_(inst.opname == "BINARY_POWER")
+            skip_power = False
+            continue
+        yield i, inst
+
+
+def walk_both_instructions(original_instructions, original_start, instructions, start):
+    """
+    Yields matching indices and instructions from the new and original instructions,
+    leaving out changes made by the sentinel transformation.
+    """
+    original_iter = islice(enumerate(original_instructions), original_start, None)
+    new_iter = non_sentinel_instructions(instructions, start)
+    inverted_comparison = False
+    while True:
+        try:
+            original_i, original_inst = next(original_iter)
+            new_i, new_inst = next(new_iter)
+        except StopIteration:
+            return
+        if (
+            inverted_comparison
+            and original_inst.opname != new_inst.opname == "UNARY_NOT"
+        ):
+            new_i, new_inst = next(new_iter)
+        inverted_comparison = (
+            original_inst.opname == new_inst.opname in ("CONTAINS_OP", "IS_OP")
+            and original_inst.arg != new_inst.arg
+        )
+        yield original_i, original_inst, new_i, new_inst
+
+
+def handle_jumps(instructions, original_instructions):
+    """
+    Transforms instructions in place until it looks more like original_instructions,
+    replacing JUMP instructions that aren't also present in original_instructions
+    with the sections that they jump to until a raise or return.
+    This is only needed in 3.10+ where optimisations lead to more drastic changes
+    after the sentinel transformation.
+    """
+    while True:
+        for original_i, original_inst, new_i, new_inst in walk_both_instructions(
+            original_instructions, 0, instructions, 0
+        ):
+            if "JUMP" in new_inst.opname and "JUMP" not in original_inst.opname:
+                # Find where the new instruction is jumping to, ignoring
+                # instructions which have been copied in previous iterations
+                start = only(
+                    i
+                    for i, inst in enumerate(instructions)
+                    if inst.offset == new_inst.argval
+                    and not getattr(inst, "_copied", False)
+                )
+                # Replace the jump instruction with the jumped to section of instructions
+                # That section may also be deleted if it's not similarly duplicated
+                # in original_instructions
+                instructions[new_i : new_i + 1] = handle_jump(
+                    original_instructions, original_i, instructions, start
+                )
+                # Start the root while loop from the beginning, checking for other jumps
+                break
+            else:
+                if not (opnames_match(original_inst, new_inst)):
+                    raise AssertionError
+        else:  # No mismatched jumps found, we're done
+            return
+
+
+def handle_jump(original_instructions, original_start, instructions, start):
+    """
+    Returns the section of instructions starting at `start` and ending
+    with a RETURN_VALUE or RAISE_VARARGS instruction.
+    There should be a matching section in original_instructions starting at original_start.
+    If that section doesn't appear elsewhere in original_instructions,
+    then also delete the returned section of instructions.
+    """
+    for original_j, original_inst, new_j, new_inst in walk_both_instructions(
+        original_instructions, original_start, instructions, start
+    ):
+        assert_(opnames_match(original_inst, new_inst))
+        if original_inst.opname in ("RETURN_VALUE", "RAISE_VARARGS"):
+            inlined = deepcopy(instructions[start : new_j + 1])
+            for inl in inlined:
+                inl._copied = True
+            orig_section = original_instructions[original_start : original_j + 1]
+            if not check_duplicates(
+                original_start, orig_section, original_instructions
+            ):
+                instructions[start : new_j + 1] = []
+            return inlined
+
+
+def check_duplicates(original_i, orig_section, original_instructions):
+    """
+    Returns True if a section of original_instructions starting somewhere other
+    than original_i and matching orig_section is found, i.e. orig_section is duplicated.
+    """
+    for dup_start in range(len(original_instructions)):
+        if dup_start == original_i:
+            continue
+        dup_section = original_instructions[dup_start : dup_start + len(orig_section)]
+        if len(dup_section) < len(orig_section):
+            return False
+        if all(
+            opnames_match(orig_inst, dup_inst) and orig_inst.lineno == dup_inst.lineno
+            for orig_inst, dup_inst in zip(orig_section, dup_section)
+        ):
+            return True
+
+
+def opnames_match(inst1, inst2):
+    return (
+        inst1.opname == inst2.opname
+        or "JUMP" in inst1.opname
+        and "JUMP" in inst2.opname
+        or (inst1.opname == "PRINT_EXPR" and inst2.opname == "POP_TOP")
+        or (
+            inst1.opname in ("LOAD_METHOD", "LOOKUP_METHOD")
+            and inst2.opname == "LOAD_ATTR"
+        )
+        or (inst1.opname == "CALL_METHOD" and inst2.opname == "CALL_FUNCTION")
+    )
+
+
 def get_setter(node):
     parent = node.parent
     for name, field in ast.iter_fields(parent):
